/***************************************************************************
 * Wrapper to the jpeglib 6b
 * should be linked with the jpeglib-6b (http://www.ijg.org/)
 * To learn how to use it, look at function TestJpeg() and jpeg2jpeg
 * 
 * Sylvain MARECHAL - sylvain.marechal1@libertysurf.fr
 **************************************************************************/
#ifdef _WIN32
#include <windows.h>
#include <io.h>
#else
#define O_BINARY    0
#endif
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <setjmp.h>
#include "dcpjpeg.h"

/*  For decompression  */
static void dcpjpeg_init_source(j_decompress_ptr cinfo);
static boolean dcpjpeg_fill_input_buffer( j_decompress_ptr cinfo );
static void dcpjpeg_skip_input_data (j_decompress_ptr cinfo, long num_bytes);
static boolean dcpjpeg_resync_to_restart( j_decompress_ptr cinfo, int desired );
static void dcpjpeg_term_source( j_decompress_ptr cinfo );

static boolean dcpjpeg_fill_input_buffer_before_one_pass( DcpJpeg * pDcpJpeg );


/*  For compression  */
static void cpjpeg_init_destination( j_compress_ptr cinfo );
static boolean cpjpeg_empty_output_buffer( j_compress_ptr cinfo );
static void cpjpeg_term_destination( j_compress_ptr cinfo );



/*  For error management  */
static void dcpjpeg_error_exit (j_common_ptr cinfo);
static void dcpjpeg_output_message (j_common_ptr cinfo);
static void dcpjpeg_emit_message (j_common_ptr cinfo, int msg_level);


#define BUFFER_REALLOC_SIZE	20480


/* Example */
typedef struct tagReadJpeg
{
    int hFile;
    char *pBuffer;
    int cbBuffer;
} ReadJpeg;

typedef struct tagWriteJpeg
{
    int hFile;
    int cb;
} WriteJpeg;


static int ReadJpegFile( void *, char ** ppBuffer, int * pcbBuffer,
        PFGetDataRelease * pGetDataRelease, void ** plParam2 );
static void ReadJpegFileRelease( void * lParam, void * lParam2 );
static int WriteJpegFile( void *, char * pBuffer, int cbBuffer );
static void ErrorMessage( void *, char * pszBuffer );



/***************************************************************************
 *  --TestJpeg--
 *
 *	Utilisation des structures DcpJpeg et CpJpeg.
 *	On decompresse 1 jpeg puis on le recompresse a la
 *	volee en changeant la qualite
 **************************************************************************/
void TestJpeg( char *pszFileSource, char *pszFileDest, int nQuality, int fOnePass )
{
    int fFirst = 1;
    ReadJpeg rj = {0};
    WriteJpeg wj = {0};
    DcpJpeg * pDcpJpeg = DcpJpegInit2( ReadJpegFile, (void * )&rj, ErrorMessage, fOnePass );
    CpJpeg * pCpJpeg = CpJpegInit( WriteJpegFile, (void * )&wj, ErrorMessage );
    
    
    DcpJpegInitScale( pDcpJpeg, 2, 1, 0 );
    
    rj.pBuffer = (char *)malloc(1024);
    rj.cbBuffer = 1024;
    if( ! rj.pBuffer )
    {
	DcpJpegDestroy( pDcpJpeg );
	CpJpegDestroy( pCpJpeg );
	return;
    }
    
    /*  Source  */
    rj.hFile = open( pszFileSource, O_RDONLY | O_BINARY, 0666 );
    if( rj.hFile == -1 )
    {
	DcpJpegDestroy( pDcpJpeg );
	CpJpegDestroy( pCpJpeg );
	return;
    }
    
    /*  iDestination  */
    wj.hFile = open( pszFileDest, O_CREAT | O_TRUNC | O_BINARY | O_RDWR , 0666 );
    if( wj.hFile == -1 )
    {
	close( rj.hFile );
	DcpJpegDestroy( pDcpJpeg );
	CpJpegDestroy( pCpJpeg );
	return;
    }
    
    
    /*  Write in destination file */
    for(;;)
    {
	int err;
	char *pDecompress; int cbDecompress;
	
        /* Uncompress data */
	err = DcpJpegDecode( pDcpJpeg, &pDecompress, &cbDecompress );
	if( cbDecompress > 0 )
	{
	    if( fFirst )
	    {
		unsigned int image_width;
		unsigned int image_height;
		int input_components;
		int in_color_space;
		
                /* Read parameters in source image */
		DcpJpegGetParams( pDcpJpeg, &image_width, &image_height, &input_components, &in_color_space );

                /* Write parameters in dest image and set the quality */
		CpJpegSetParams( pCpJpeg, image_width, image_height, input_components, in_color_space, nQuality );
		fFirst = 0;
	    }

            /* Compress the image */
	    CpJpegEncode( pCpJpeg, pDecompress, cbDecompress );
	}
	if( err < 0 )
        {
            /* Err < 0 means end or error */
	    break;
        }
    }
    
    
    /* free memory and closes file */
    DcpJpegDestroy( pDcpJpeg );
    CpJpegDestroy( pCpJpeg );
    close( rj.hFile );
    free( rj.pBuffer );
    close( wj.hFile );
}

/***************************************************************************
 *  --ReadJpegFile--
 *
 **************************************************************************/
int ReadJpegFile( void * lParam, char **ppBuffer, int *pcbBuffer,
        PFGetDataRelease * pGetDataRelease, void ** plParam2 )
{
    ReadJpeg *prj = (ReadJpeg *)lParam;
    int nRet;
    static int b = 0;
    *pGetDataRelease = ReadJpegFileRelease;
    *plParam2 = NULL;
    
    if( b )
    {
	b = 0;
	*ppBuffer = NULL;
	*pcbBuffer = 0;
	return 0;
    }
    b = 1;
    
    nRet = read( prj->hFile, prj->pBuffer, prj->cbBuffer );
    *ppBuffer = (nRet <= 0 ? NULL : prj->pBuffer);
    *pcbBuffer = (nRet <= 0 ? 0 : nRet);
    
    if( nRet <= 0 )
    {
	nRet = -1;
    }
    
    return nRet;
}


/***************************************************************************
 *  --ReadJpegFileRelease--
 *
 **************************************************************************/
void ReadJpegFileRelease( void * lParam, void * lParam2 )
{
}

/***************************************************************************
 *  --WriteJpegFile--
 *
 **************************************************************************/
int WriteJpegFile( void * lParam, char * pBuffer, int cbBuffer )
{
    WriteJpeg *pwj = (WriteJpeg*)lParam;    

    if( pBuffer && cbBuffer )
    {
        write( pwj->hFile, pBuffer, cbBuffer );
        pwj->cb += cbBuffer;
    }
    else
    {
        printf( "Size of file : %d\n", pwj->cb );
    }
    return 0;
}

/***************************************************************************
 *  --ErrorMessage--
 *
 **************************************************************************/
void ErrorMessage( void * lParam, char * pszBuffer )
{
    printf( pszBuffer );
}

/**************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************/

/***************************************************************************
 * --DcpJpegInit-- fOnePass = 0
 * --DcpJpegInit2-- idem avec fOnePass = 0 ou 1
 *
 *  fOnePass etait cence ameliorer les perfomances,
 *  en fournissant toute la donnee a decompresser en
 *  une seule fois, mais cela ne semble pas marcher
 *  plus vite, au contraire ...
 *
 **************************************************************************/
DcpJpeg * DcpJpegInit( PFGetData pfGetData, void * lParam,
	PFErrorMessage pfErrorMessage )
{
    return DcpJpegInit2( pfGetData, lParam, pfErrorMessage, 0  ) ;
}
DcpJpeg * DcpJpegInit2( PFGetData pfGetData, void * lParam,
	PFErrorMessage pfErrorMessage, int fOnePass  )
{
    DcpJpeg * pDcpJpeg;
    struct jpeg_error_mgr *pjerr;
    struct jpeg_decompress_struct * pdjpeg;
    
    /*  Alloc structure  */
    pDcpJpeg = (DcpJpeg *)calloc( 1, sizeof(DcpJpeg) );
    if( ! pDcpJpeg ) return pDcpJpeg;
    
    /*  Rebouclage sur structure mere  */
    pDcpJpeg->jerrext.UnionCpDcp.pDcpJpeg = pDcpJpeg;
    pDcpJpeg->djpegext.pDcpJpeg = pDcpJpeg;
    
    /*  Gestion des erreurs generique  */
    pDcpJpeg->jerrext.lParam = lParam;
    pDcpJpeg->jerrext.ErrorMessage = pfErrorMessage;
    
    /*  Variables plus courtes  */
    pjerr = &pDcpJpeg->jerrext.jerr;
    pdjpeg = &pDcpJpeg->djpegext.djpeg;
    
    /*  init jpeg err */
    jpeg_std_error( pjerr);
    pjerr->error_exit     = dcpjpeg_error_exit ;
    pjerr->emit_message   = dcpjpeg_emit_message ;
    pjerr->output_message = dcpjpeg_output_message ;
    pjerr->trace_level = 0 ; /* pas de tracing (0) */
    pdjpeg->err = pjerr ;
    
    /*  init jpeg decompression */
    jpeg_create_decompress( pdjpeg) ;
    pdjpeg->src = (struct jpeg_source_mgr *)malloc( sizeof(struct jpeg_source_mgr) ) ;
    pdjpeg->src->next_input_byte   = NULL ;
    pdjpeg->src->bytes_in_buffer    = 0 ;
    pdjpeg->src->init_source       = dcpjpeg_init_source ;
    pdjpeg->src->fill_input_buffer = dcpjpeg_fill_input_buffer ;
    pdjpeg->src->skip_input_data   = dcpjpeg_skip_input_data ;
    pdjpeg->src->resync_to_restart = dcpjpeg_resync_to_restart ;
    pdjpeg->src->term_source       = dcpjpeg_term_source ;
    
    
    /*  Init reste de la structure  */
    pDcpJpeg->nState = 0;
    pDcpJpeg->GetData = pfGetData;
    pDcpJpeg->lParam = lParam;
    
    
    /*  Decompression en une passe eventuellement */
    pDcpJpeg->fOnePass = fOnePass;
    return pDcpJpeg;
}

/***************************************************************************
 *  --DcpJpegInitScale--
 *
 *  Fonction a appeler si on veut modifier la taille de l'image
 *  en sortie. Soit on fournit scale_num et scale_denom qui sont le
 *  numerateur et le denom (par ex, 1 image a 1/8)
 *  soit on fournit output_width qui est la largeur approximative
 *  desiree
 *
 **************************************************************************/
void DcpJpegInitScale( DcpJpeg * pDcpJpeg, int scale_num, int scale_denom, int output_width )
{
    pDcpJpeg->scale_num = scale_num;
    pDcpJpeg->scale_denom = scale_denom;
    pDcpJpeg->output_width = output_width;
}

/***************************************************************************
 *  --DcpJpegDestroy--
 *
 **************************************************************************/
void DcpJpegDestroy( DcpJpeg * pDcpJpeg )
{
    if( ! pDcpJpeg ) return;
    if( pDcpJpeg->djpegext.djpeg.src )	free(pDcpJpeg->djpegext.djpeg.src);
    jpeg_destroy_decompress(&pDcpJpeg->djpegext.djpeg);
    if( pDcpJpeg->pBuffer ) free( pDcpJpeg->pBuffer );
    if( pDcpJpeg->pSaveBuffer ) free( pDcpJpeg->pSaveBuffer );
    free( pDcpJpeg );
}

/***************************************************************************
 *  --DcpJpegGetParams--
 *
 **************************************************************************/
void DcpJpegGetParams( DcpJpeg * pDcpJpeg, unsigned int * pimage_width,
		     unsigned int * pimage_height,int * pinput_components, 
		     int * pin_color_space )
{
    struct jpeg_decompress_struct * pdjpeg;
    
    /*  Variable plus courte  */
    pdjpeg = &pDcpJpeg->djpegext.djpeg;
    
    *pimage_width  = pdjpeg->output_width;
    *pimage_height = pdjpeg->output_height;
    *pinput_components = pdjpeg->output_components;
    *pin_color_space = pdjpeg->out_color_space;
}

/***************************************************************************
 *  --DcpJpegDecode--
 *
 * Uncompresses JPEG
 *
 **************************************************************************/
int DcpJpegDecode( DcpJpeg * pDcpJpeg, char **ppDecompress, int * pcbDecompress )
{
    struct jpeg_decompress_struct * pdjpeg;
    
    /*  inits  */
    *ppDecompress = NULL;
    *pcbDecompress = 0;
    
    /*  Gestion de la decompression en une passe */
    if( pDcpJpeg->fOnePass )
    {
	if( ! dcpjpeg_fill_input_buffer_before_one_pass( pDcpJpeg ) )
	    return 0;
    }
    
    /*  Gestion des erreurs  */
    if( setjmp( pDcpJpeg->jerrext.jmp_env ) != 0 )
    {
	return -2;
    }
    
    /*  Variables plus courtes  */
    pdjpeg = &pDcpJpeg->djpegext.djpeg;
    
    switch(pDcpJpeg->nState)
    {
    case 0:
	{
	    /*  on commence par decoder l'entete  */
	    if( jpeg_read_header( pdjpeg, 1 ) == JPEG_SUSPENDED )
		return 0 ;
	    pDcpJpeg->nState = 1;
	}
    case 1:
	{
	    /*  C'est ici qu'on prend en compte l'echelle  */
	    if( pDcpJpeg->scale_num && pDcpJpeg->scale_denom )
	    {
		pdjpeg->scale_num = pDcpJpeg->scale_num;
		pdjpeg->scale_denom = pDcpJpeg->scale_denom;
	    }
	    /*  Dans le cas ou on donne la largeur, on calcule une echelle
	    (tres) approximative  */
	    else if( pDcpJpeg->output_width )
	    {
		if( pDcpJpeg->output_width > (int)pdjpeg->image_width )
		{
		    pdjpeg->scale_num = pDcpJpeg->output_width / pdjpeg->image_width;
		    pdjpeg->scale_denom = 1;
		}
		else
		{
		    pdjpeg->scale_num = 1;
		    pdjpeg->scale_denom = pdjpeg->image_width / pDcpJpeg->output_width;
		}
	    }
	    
	    /*  init decompression  */
	    if( jpeg_start_decompress( pdjpeg ) == 0 )
		return 0 ;
		/*  pour info, les parametres suivants sont a jour :
		output_width		image width and height, as scaled
		output_height
		out_color_components	# of color components in out_color_space
		output_components	# of color components returned per pixel
		colormap		the selected colormap, if any
		actual_number_of_colors		number of entries in colormap
	    */
	    pDcpJpeg->pBuffer = (char *)malloc( (3*pdjpeg->output_width) * (pDcpJpeg->fOnePass ? pdjpeg->output_height + 1 : 1) );
	    pDcpJpeg->cbBuffer = (3*pdjpeg->output_width) * (pDcpJpeg->fOnePass ? pdjpeg->output_height + 1: 1);
	    pDcpJpeg->nState = 2;
	}
    case 2:
	{
	    /*  Tant qu'il y a des lignes  */
	    while( pdjpeg->output_scanline < pdjpeg->output_height ) 
	    {
		JSAMPROW pRow[1];
		
		/*  On lit une ligne a la fois  */
		pRow[0] = pDcpJpeg->pBuffer + (*pcbDecompress);
		if( jpeg_read_scanlines( pdjpeg, pRow, 1) == 0 )
		{
		    *ppDecompress = pDcpJpeg->pBuffer;
		    return 0;
		}
		/*  On doit reallouer  */
		if( pDcpJpeg->cbBuffer == (int)((*pcbDecompress) + (3*pdjpeg->output_width)) )
		{
		    char * pSave;
		    
		    pSave = pDcpJpeg->pBuffer;
		    pDcpJpeg->pBuffer = (char *)realloc(pDcpJpeg->pBuffer, pDcpJpeg->cbBuffer + (3*pdjpeg->output_width) );
		    if( pDcpJpeg->pBuffer == NULL )
		    {
			pDcpJpeg->pBuffer = pSave;
			*ppDecompress = pDcpJpeg->pBuffer;
			return 0;
		    }
		    pDcpJpeg->cbBuffer += (3*pdjpeg->output_width);
		}
		/*  On incremente le nombre d'octets lus  */
		*pcbDecompress += (3*pdjpeg->output_width);
	    }
	    *ppDecompress = pDcpJpeg->pBuffer;
	    pDcpJpeg->nState = 3;
	}
    case 3:
	{
	    if( !jpeg_finish_decompress( pdjpeg ) )
		return 0 ;
	    pDcpJpeg->nState = 4;
	}
    case 4:
	{
	    return -1;		    
	}
    }
    return -2;
}

/***************************************************************************
 *  --dcpjpeg_init_source--
 *
 **************************************************************************/
void dcpjpeg_init_source(j_decompress_ptr cinfo)
{
    /*  Rien a faire  */
}

/***************************************************************************
 *  --dcpjpeg_fill_input_buffer_before_one_pass--
 *
 **************************************************************************/
boolean dcpjpeg_fill_input_buffer_before_one_pass( DcpJpeg * pDcpJpeg )
{
    int nRead;
    char * pBuffer; int cbBuffer;
    PFGetDataRelease GetDataRelease = NULL;
    void * lParam2 = NULL;
    int ret;
    
    /*  Lecture des donnees : > 0 -> ok, 0 -> rien a lire, 
    -1 -> fini, -2 -> erreur  */
    nRead = (*pDcpJpeg->GetData)(pDcpJpeg->lParam, &pBuffer, &cbBuffer, &GetDataRelease, &lParam2 );
    if( nRead == 0 )
    {
	/*  On a rien a lire */
	ret = 0;
    }
    else if( nRead < 0 )
    {
	/*  Y'a plus de donnees : C'est fini */
	ret = 1;
    }
    else
    {		
	/*  On realloc le buffer */
	if( pDcpJpeg->cbSaveBuffer < cbBuffer + pDcpJpeg->cbSaveLast )
	{
	    char *pSave = pDcpJpeg->pSaveBuffer;
	    int cbRealloc = ( cbBuffer < BUFFER_REALLOC_SIZE ? BUFFER_REALLOC_SIZE : cbBuffer);
	    
	    pDcpJpeg->pSaveBuffer = (char *)realloc( pDcpJpeg->pSaveBuffer, pDcpJpeg->cbSaveLast + cbRealloc );
	    if( pDcpJpeg->pSaveBuffer == NULL )
	    {
		pDcpJpeg->pSaveBuffer = pSave;
                if( GetDataRelease ) GetDataRelease( pDcpJpeg->lParam, lParam2 );
		return 1;	/*kk*/
	    }
	    pDcpJpeg->cbSaveBuffer = pDcpJpeg->cbSaveLast + cbRealloc;
	}
	
	/*  On sauvegarde les donnees */
	memcpy( pDcpJpeg->pSaveBuffer + pDcpJpeg->cbSaveLast, pBuffer, cbBuffer );
	pDcpJpeg->cbSaveLast = cbBuffer + pDcpJpeg->cbSaveLast;
	ret = 0;
    }

    /* Dat is copied */
    if( GetDataRelease ) GetDataRelease( pDcpJpeg->lParam, lParam2 );
    return ret;
}    

/***************************************************************************
 *  --dcpjpeg_fill_input_buffer_one_pass--
 *
 **************************************************************************/
boolean dcpjpeg_fill_input_buffer_one_pass( j_decompress_ptr cinfo )
{
    jpeg_dc_struct_ext * jExt = (jpeg_dc_struct_ext *)cinfo;
    DcpJpeg * pDcpJpeg = jExt->pDcpJpeg;
    static unsigned char jpeg_eoi[2] = {0xff,JPEG_EOI};

    if( pDcpJpeg->fDataGiven )
    {
	/*  Y'a plus de donnees : On indique que c'est fini */
	cinfo->src->next_input_byte = jpeg_eoi;
	cinfo->src->bytes_in_buffer = 2;
	
	return 1;
    }
    pDcpJpeg->fDataGiven = 1;
    cinfo->src->next_input_byte = pDcpJpeg->pSaveBuffer;
    cinfo->src->bytes_in_buffer = pDcpJpeg->cbSaveBuffer;
    return 1;
}

/***************************************************************************
 *  --dcpjpeg_fill_input_buffer--
 *
 **************************************************************************/
boolean dcpjpeg_fill_input_buffer( j_decompress_ptr cinfo )
{
    jpeg_dc_struct_ext * jExt = (jpeg_dc_struct_ext *)cinfo;
    DcpJpeg * pDcpJpeg = jExt->pDcpJpeg;
    int nRead, nRet;
    char * pBuffer; int cbBuffer;
    PFGetDataRelease GetDataRelease = NULL;
    void * lParam2 = NULL;
    static unsigned char jpeg_eoi[2] = {0xff,JPEG_EOI};

    /* Gestion de la decompression en une passe */
    if( jExt->pDcpJpeg->fOnePass )
	return dcpjpeg_fill_input_buffer_one_pass( cinfo );
    
    
	
    /*  Lecture des donnees : > 0 -> ok, 0 -> rien a lire, 
    -1 -> fini, -2 -> erreur  */
    nRead = (*pDcpJpeg->GetData)(pDcpJpeg->lParam, &pBuffer, &cbBuffer, &GetDataRelease, &lParam2 );
    if( nRead == 0 )
    {
    /*  On a rien a lire : la librairie nous indique de combien
	elle va reculer au prochain appel . On doit le memoriser */
	pDcpJpeg->nOffset = cinfo->src->bytes_in_buffer;
	cinfo->src->next_input_byte = NULL ;
	cinfo->src->bytes_in_buffer = 0 ;
	
	nRet = 0;
    }
    else if( nRead < 0 )
    {
	/*  Y'a plus de donnees : On indique que c'est fini */
	cinfo->src->next_input_byte = jpeg_eoi;
	cinfo->src->bytes_in_buffer = 2;
	
	nRet = 1;
    }
    else
    {		
	/*  On commence par decaler la partie jusqu'au marker de depart  */
	if( pDcpJpeg->nOffset )
	    memcpy( pDcpJpeg->pSaveBuffer, 
	    pDcpJpeg->pSaveBuffer + pDcpJpeg->cbSaveLast - pDcpJpeg->nOffset, 
	    pDcpJpeg->nOffset ); 
	
	/*  On realloc le buffer s'il est trop petit  */
	if( pDcpJpeg->cbSaveBuffer < cbBuffer + pDcpJpeg->nOffset )
	{
	    char *pSave = pDcpJpeg->pSaveBuffer;
	    
	    pDcpJpeg->pSaveBuffer = (char *)realloc( pDcpJpeg->pSaveBuffer, cbBuffer + pDcpJpeg->nOffset );
	    if( pDcpJpeg->pSaveBuffer == NULL )
	    {
		pDcpJpeg->pSaveBuffer = pSave;
		cinfo->src->next_input_byte = jpeg_eoi;
		cinfo->src->bytes_in_buffer = 2;
                if( GetDataRelease ) GetDataRelease( pDcpJpeg->lParam, lParam2 );
		return 1;	/*kk*/
	    }
	    pDcpJpeg->cbSaveBuffer = cbBuffer + pDcpJpeg->nOffset;
	}
	
	/*  On sauvegarde les donnees pour le prochain appel  */
	memcpy( pDcpJpeg->pSaveBuffer + pDcpJpeg->nOffset, pBuffer, cbBuffer );
	
	/*  On passe l'info a la jpeg lib  */
	cinfo->src->next_input_byte = pDcpJpeg->pSaveBuffer;
	cinfo->src->bytes_in_buffer = cbBuffer + pDcpJpeg->nOffset;
	
	/*  On memorise la quantite de donnees  */
	pDcpJpeg->cbSaveLast = cbBuffer + pDcpJpeg->nOffset;
	pDcpJpeg->nOffset = 0;
	
	nRet = 1;
	
	/*  Gestion des donnees a sauter  */
	if( pDcpJpeg->cbToSkip > 0 )
	{
	    dcpjpeg_skip_input_data( cinfo, pDcpJpeg->cbToSkip );
	    if( pDcpJpeg->cbToSkip > 0 || cinfo->src->bytes_in_buffer == 0 )
		nRet = 0;
	}
    }
    

    /* Dat is copied */
    if( GetDataRelease ) GetDataRelease( pDcpJpeg->lParam, lParam2 );
    return nRet;
}

/***************************************************************************
 *  --dcpjpeg_skip_input_data--
 *
 **************************************************************************/
void dcpjpeg_skip_input_data( j_decompress_ptr cinfo, long num_bytes)
{
    jpeg_dc_struct_ext * jExt = (jpeg_dc_struct_ext *)cinfo;
    DcpJpeg * pDcpJpeg = jExt->pDcpJpeg;
    
    /* a-t-on suffisament de data en buffer pour faire completement 
    ** le skip ici ?
    */
    if( (long)cinfo->src->bytes_in_buffer > num_bytes )
    {
        cinfo->src->next_input_byte += num_bytes;
        cinfo->src->bytes_in_buffer -= num_bytes;
        pDcpJpeg->cbToSkip = 0;
    }
    else
    {
    /* pas assez de place en buffer : on conserve le nb de bytes a skipper
    ** pour le prochain appel a 'zjpeg_fill_input_buffer'
        */
        pDcpJpeg->cbToSkip = num_bytes - cinfo->src->bytes_in_buffer;
        cinfo->src->next_input_byte = NULL;
        cinfo->src->bytes_in_buffer  = 0;
    }
}

/***************************************************************************
 *  --dcpjpeg_resync_to_restart--
 *
 **************************************************************************/
boolean dcpjpeg_resync_to_restart( j_decompress_ptr cinfo, int desired )
{
    return jpeg_resync_to_restart( cinfo, desired );
}

/***************************************************************************
 *  --dcpjpeg_term_source--
 *
 **************************************************************************/
void dcpjpeg_term_source( j_decompress_ptr cinfo )
{
    /*  Rien a faire  */
}

/**************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************
 **************************************************************************/

/***************************************************************************
 *  --CpJpegInit--
 *
 **************************************************************************/
CpJpeg * CpJpegInit( PFWriteData pfWriteData, void * lParam, PFErrorMessage pfErrorMessage  )
{
    CpJpeg * pCpJpeg;
    struct jpeg_error_mgr *pjerr;
    struct jpeg_compress_struct * pcjpeg;
    
    /*  Alloc structure  */
    pCpJpeg = (CpJpeg *)calloc( 1, sizeof(CpJpeg) );
    if( ! pCpJpeg ) return pCpJpeg;
    
    /*  Rebouclage sur structure mere  */
    pCpJpeg->jerrext.UnionCpDcp.pCpJpeg = pCpJpeg;
    pCpJpeg->cjpegext.pCpJpeg = pCpJpeg;
    
    /*  Variables plus courtes  */
    pjerr = &pCpJpeg->jerrext.jerr;
    pcjpeg = &pCpJpeg->cjpegext.cjpeg;
    
    /*  Gestion des erreurs generique  */
    pCpJpeg->jerrext.lParam = lParam;
    pCpJpeg->jerrext.ErrorMessage = pfErrorMessage;
    
    /*  init jpeg err */
    jpeg_std_error( pjerr);
    pjerr->error_exit     = dcpjpeg_error_exit ;
    pjerr->emit_message   = dcpjpeg_emit_message ;
    pjerr->output_message = dcpjpeg_output_message ;
    pjerr->trace_level = 0 ; /* pas de tracing (0) */
    pcjpeg->err = pjerr ;
    
    /*  init jpeg decompression */
    jpeg_create_compress( pcjpeg) ;
    pcjpeg->dest = (struct jpeg_destination_mgr *)malloc( sizeof(struct jpeg_destination_mgr) ) ;
    pcjpeg->dest->init_destination   = cpjpeg_init_destination;
    pcjpeg->dest->empty_output_buffer= cpjpeg_empty_output_buffer;
    pcjpeg->dest->term_destination   = cpjpeg_term_destination;
    
    /*  Init reste de la structure  */
    pCpJpeg->nState = 0;
    pCpJpeg->WriteData = pfWriteData;
    pCpJpeg->lParam = lParam;
    pCpJpeg->pBuffer = (char *)malloc( BUFFER_REALLOC_SIZE );	/**/
    if( pCpJpeg->pBuffer == NULL )
    {
	CpJpegDestroy( pCpJpeg );
	return NULL;
    }
    pCpJpeg->cbBuffer = BUFFER_REALLOC_SIZE;
    return pCpJpeg;
}

/***************************************************************************
 *  --CpJpegDestroy--
 *
 **************************************************************************/
void CpJpegDestroy( CpJpeg * pCpJpeg )
{
    struct jpeg_compress_struct * pcjpeg;
    
    /*  Test preliminaire  */
    if( ! pCpJpeg ) return;
    
    /*  On termine eventuellement  */
    CpJpegEncode( pCpJpeg, NULL, 0 );
    
    /*  Variable plus courte  */
    pcjpeg = &pCpJpeg->cjpegext.cjpeg;
    
    if( pcjpeg->dest )	free(pcjpeg->dest);
    jpeg_destroy_compress(pcjpeg);
    if( pCpJpeg->pBuffer ) free(pCpJpeg->pBuffer);
    free( pCpJpeg );
}

/***************************************************************************
 *  --CpJpegSetParams--
 *
 **************************************************************************/
void CpJpegSetParams( CpJpeg * pCpJpeg, unsigned int image_width, unsigned int image_height,int input_components, int in_color_space, int nQuality )
{
    struct jpeg_compress_struct * pcjpeg;
    
    /*  Variable plus courte  */
    pcjpeg = &pCpJpeg->cjpegext.cjpeg;
    
    pcjpeg->image_width  = image_width;
    pcjpeg->image_height = image_height;
    pcjpeg->input_components = input_components;
    pcjpeg->in_color_space = in_color_space;
    
    pCpJpeg->nQuality = nQuality;
}

/***************************************************************************
 *  --CpJpegEncode--
 *
 **************************************************************************/
int CpJpegEncode( CpJpeg * pCpJpeg, char *pData, int cbData )
{
    struct jpeg_compress_struct * pcjpeg;
    
    /*  Test preliminaire  */
    if( pData == NULL || cbData == 0 )
    {
	if( pCpJpeg->nState != 1 )
	    return 0;
	/*  Pour finir  */
	pCpJpeg->nState = 2;
    }
    
    /*  Variable plus courte  */
    pcjpeg = &pCpJpeg->cjpegext.cjpeg;
    
    /*  Gestion des erreurs  */
    if( setjmp( pCpJpeg->jerrext.jmp_env ) != 0 )
    {
	return -2;
    }
    
    /*  Compression du jpeg  */
    switch(pCpJpeg->nState)
    {
    case 0:
	{
	    jpeg_set_defaults( pcjpeg );
	    jpeg_set_quality( pcjpeg, pCpJpeg->nQuality, 1);
	    pcjpeg->optimize_coding = 1;	    
	    jpeg_simple_progression(pcjpeg); 
	    jpeg_start_compress( pcjpeg, 1 );
	    pCpJpeg->nState = 1;
	}
    case 1:
	{
	    JSAMPROW pRow[1];
	    
	    /*  Test preliminaire  */
	    if( cbData % (pcjpeg->image_width*3) != 0 )
		printf( "Partial line ...\n");
	    
		/*  On consomme toute la donnee passee en argument. Celle-ci doit
	    etre multiple du nombre de lignes  */
	    while( 1 )
	    {
		pRow[0] = pData;
		jpeg_write_scanlines( pcjpeg, pRow, 1) ;
		pCpJpeg->nLineDones ++;
		pData += (pcjpeg->image_width*3);
		cbData -= (pcjpeg->image_width*3);
		if( pCpJpeg->nLineDones == (int)(pcjpeg->image_height) )
		{
		    break;
		}
		else if( cbData <= 0 )
		{
		    return 0;
		}
	    }
	    pCpJpeg->nState = 2;
	}
    case 2:
	{
	    jpeg_finish_compress(pcjpeg);
	    pCpJpeg->nState = 3;
	}
    case 3:
	{
	}
    }
    return 0;
}

/***************************************************************************
 *  --cpjpeg_init_destination--
 *
 **************************************************************************/
void cpjpeg_init_destination( j_compress_ptr cinfo )
{
    jpeg_cp_struct_ext * jExt = (jpeg_cp_struct_ext *)cinfo;
    CpJpeg * pCpJpeg = jExt->pCpJpeg;
    
    /*  Init buffer  */
    cinfo->dest->next_output_byte   = pCpJpeg->pBuffer;
    cinfo->dest->free_in_buffer     = pCpJpeg->cbBuffer;
}

/***************************************************************************
 * --cpjpeg_empty_output_buffer--
 *
 *  called when buffer should be emptied
 *
 **************************************************************************/
boolean cpjpeg_empty_output_buffer( j_compress_ptr cinfo )
{
    jpeg_cp_struct_ext * jExt = (jpeg_cp_struct_ext *)cinfo;
    CpJpeg * pCpJpeg = jExt->pCpJpeg;
    
    /*  On appelle la fonction  */
    (*pCpJpeg->WriteData)( pCpJpeg->lParam, pCpJpeg->pBuffer, pCpJpeg->cbBuffer );
    
    cinfo->dest->next_output_byte   = pCpJpeg->pBuffer;
    cinfo->dest->free_in_buffer   = pCpJpeg->cbBuffer;
    return 1;
}

/***************************************************************************
 *  --cpjpeg_term_destination--
 *
 *  Called when compression is ended
 *
 **************************************************************************/
void cpjpeg_term_destination( j_compress_ptr cinfo )
{
    jpeg_cp_struct_ext * jExt = (jpeg_cp_struct_ext *)cinfo;
    CpJpeg * pCpJpeg = jExt->pCpJpeg;
    
    /*  On appelle la fonction  */
    (*pCpJpeg->WriteData)( pCpJpeg->lParam, pCpJpeg->pBuffer, 
	pCpJpeg->cbBuffer - cinfo->dest->free_in_buffer );
    (*pCpJpeg->WriteData)( pCpJpeg->lParam, NULL, 0 );
}

/***************************************************************************
 *  --dcpjpeg_error_exit--
 *
 **************************************************************************/
void dcpjpeg_error_exit(j_common_ptr cinfo)
{
    struct jpeg_error_mgr * err = cinfo->err;
    jpeg_error_mgr_ext * jerrExt = (jpeg_error_mgr_ext *)err;
    
    if( jerrExt->ErrorMessage )
	(*jerrExt->ErrorMessage)(jerrExt->lParam, "JPEG: fatal error. Exiting.\n" );
    longjmp( jerrExt->jmp_env, 1 );
}

/***************************************************************************
 *  --dcpjpeg_output_message--
 *
 **************************************************************************/
void dcpjpeg_output_message(j_common_ptr cinfo)
{
    struct jpeg_error_mgr * err = cinfo->err;
    jpeg_error_mgr_ext * jerrExt = (jpeg_error_mgr_ext *)err;
    char buffer[JMSG_LENGTH_MAX];

    /* Create the message */
    (*cinfo->err->format_message) (cinfo, buffer);

    /* Send it to stderr, adding a newline */
    if( jerrExt->ErrorMessage )
	(*jerrExt->ErrorMessage)(jerrExt->lParam, buffer );
}

/***************************************************************************
 *  --dcpjpeg_emit_message--
 *
 **************************************************************************/
void dcpjpeg_emit_message(j_common_ptr cinfo, int msg_level)
{
  struct jpeg_error_mgr * err = cinfo->err;

  if (msg_level < 0) {
    /* It's a warning message.  Since corrupt files may generate many warnings,
     * the policy implemented here is to show only the first warning,
     * unless trace_level >= 3.
     */
    if (err->num_warnings == 0 || err->trace_level >= 3)
      (*err->output_message) (cinfo);
    /* Always count warnings in num_warnings. */
    err->num_warnings++;
  } else {
    /* It's a trace message.  Show it if trace_level >= msg_level. */
    if (err->trace_level >= msg_level)
      (*err->output_message) (cinfo);
  }
}

