/*!
 * \file
 * \author Sylvain MARECHAL
 * \date 23/04/2005
 *
 * \brief Simplified JPEG API
 *
 * This API uses the JPEG library - http://www.ijg.org/
 *          
 * $Id$
 */

#ifndef __DCPJPEG_H__
#define __DCPJPEG_H__

/* #includes */
#include <setjmp.h>
#include <jpeglib.h>


#ifdef __cplusplus
extern "C" {
#endif

#ifdef _WIN32
typedef struct DcpJpeg;
typedef struct CpJpeg;
#endif

/*!\brief Optional callback given to the read callback
 *
 * \param lParam User data given in the DcpJpegInit call
 * \param lParam2 User data2 given in the PFGetdata call
 */
typedef void (*PFGetDataRelease)( void * lParam, void * lParam2);

/*!\brief Read callback 
 *
 * Note that the JPEG lib will copy data, so, try to avoid data 
 * copy in the Getdata() function and use the GetDataRelease
 * call if necessary. 
 * The GetDataRelease is called when data is copied
 *
 * \param lParam [IN] User data given in the DcpJpegInit call
 * \param ppBuffer  [OUT] pointer to data
 * \param pcbBuffer [OUT] size of data
 * \param pGetDataRelease [OUT] callback called when data is copied
 * \param plParam2 [OUT] Data given to GetDataRelease
 * \return Qtity of data readen, -1 when end of stream
 */
typedef int (*PFGetData)( void * lParam, char ** ppBuffer, int * pcbBuffer, 
        PFGetDataRelease * pGetDataRelease, void ** plParam2 );

/*!\brief Write callback 
 *
 * \param lParam [IN] User data given in the CpJpegInit call
 * \pBuffer Buffer to write
 * \cbBuffer Size of buffer
 */
typedef int (*PFWriteData)(void * lParam, char * pBuffer, int cbBuffer);

/*\brief Error callback 
 *
 * \param lParam [IN] User data
 * \param pszBuffer [IN] error message
 */
typedef void (*PFErrorMessage)( void * lParam, char * pszBuffer );

/*
typedef void * (*PFMalloc)(void * lParam, int cb );
typedef void * (*PFCalloc)(void * lParam, int cb );
typedef void * (*PFRealloc)(void * lParam, void * pointer, int cb );
typedef void   (*PFFree)( void * lParam, void * pointer );
*/

/*!
 * \brief Extended structures given to the JPEG library
 *
 * The important thing to remember is tu put the JPEG structure
 * in first position.
 * This structure is used with decompression functions.
 */
typedef struct jpeg_dc_struct_ext
{
    /*!\brief The jpeg decompression structure */
    struct jpeg_decompress_struct djpeg;        

    /*!\brief Our private data */
    struct DcpJpeg *pDcpJpeg;
} jpeg_dc_struct_ext;

/*!
 * \brief Extended structures given to the JPEG library
 *
 * The important thing to remember is tu put the JPEG structure
 * in first position.
 * This structure is used with compression functions.
 */
typedef struct jpeg_cp_struct_ext	/* Compression  */
{
    /*!\brief The jpeg compression structure */
    struct jpeg_compress_struct cjpeg;
    
    /*!\brief Our private data */
    struct CpJpeg *pCpJpeg;
} jpeg_cp_struct_ext;


/*!
 * \brief Extended structures given to the JPEG library
 *
 * The important thing to remember is tu put the JPEG structure
 * in first position.
 * This structure is used with compression and decompression 
 * functions to handle errors.
 */
typedef struct jpeg_error_mgr_ext
{
    /*!\brief The jpeg error structure */
    struct jpeg_error_mgr jerr;

    /*!\brief setjmp / longjmp to manage exceptions */
    jmp_buf	jmp_env;

    /*!\brief User param */
    void * lParam;
    PFErrorMessage ErrorMessage;

    union {
	struct DcpJpeg *pDcpJpeg;
	struct CpJpeg *pCpJpeg;
    } UnionCpDcp;

} jpeg_error_mgr_ext;


/* Structure gerant la decompression  */
typedef struct DcpJpeg
{
    jpeg_dc_struct_ext djpegext;
    jpeg_error_mgr_ext jerrext;
    int	nState;

    char	*pBuffer;	/* Donnees decompressees  */
    int	cbBuffer;	/* Taille du buffer	*/

    char	*pSaveBuffer;	/* Buffer de sauvegarde  */
    int	cbSaveBuffer;	/* Taille du buffer	*/
    int	cbSaveLast;	/* Quantite actuelle	*/
    int	nOffset;	/* = recul a effectuer	*/
    int	cbToSkip;

    int	fOnePass;	/* Decompression en une passe */
    int	fDataGiven;	/* Donnees deja donnees */

    /* Gestion chgt d'echelle (soit num et denom, soit output, pas les 2)*/
    int scale_num;		/* Numerateur pour l'echelle  */
    int scale_denom;	/* Denominateur pour l'echelle */
    int output_width;	/* Largeur desiree */


    void * lParam;		/* En general, pointeur sur structure mere  */
    PFGetData GetData;	/* Fonction utilisee pour extraire les donnees  */
} DcpJpeg;



/* Structure gerant la compression  */
typedef struct CpJpeg
{
    jpeg_cp_struct_ext cjpegext;
    jpeg_error_mgr_ext jerrext;
    int	nState;

    char	*pBuffer;	/* Donnees compressees  */
    int	cbBuffer;	/* Taille du buffer	 */
    int	nLineDones;	/* Nombre de lignes ok  */

    int nQuality;		/* Qualite de 0 a 100   */

    void * lParam;		/* En general, pointeur sur structure mere  */
    PFWriteData WriteData;	/* Fonction utilisee pour ecrire les donnees  */

    jmp_buf	jmp_env;	/* Gestion des erreurs */
} CpJpeg;


/*!\brief Creates the object used to handle decompression
 *
 * \param GetData [IN] Callback used to read the data
 * \param lParam  [IN] User data given to GetData and ErrorMessage
 * \param ErrorMessage [IN] Callback called when error
 * \return Pointer to a DcpJpeg object
 */
DcpJpeg * DcpJpegInit( PFGetData GetData, void * lParam, PFErrorMessage ErrorMessage );

/*!\brief Creates the object used to handle decompression
 *
 * \param GetData [IN] Callback used to read the data
 * \param lParam  [IN] User data given to GetData and ErrorMessage
 * \param ErrorMessage [IN] Callback called when error
 * \param fOnePass [IN] If true, do the decompression in one pass
 * \return Pointer to a DcpJpeg object
 */
DcpJpeg * DcpJpegInit2( PFGetData GetData, void * lParam, PFErrorMessage ErrorMessage, int fOnePass );

/*!\brief Sets the scale
 *
 * \param pDcpJpeg [IN] Pointer to object created with DcpJpegInit(2) functions
 * \param scale_num
 * \param scale_denom
 * \param output_width
 */
void DcpJpegInitScale( DcpJpeg * pDcpJpeg, int scale_num, int scale_denom, int output_width );

/*!\brief Destroys the object
 * 
 * \param pDcpJpeg [IN]Pointer to object created with DcpJpegInit(2) functions
 */
void DcpJpegDestroy( DcpJpeg * pDcpJpeg );

/*!\brief Returns informations for current image
 *
 * \param pDcpJpeg [IN] Pointer to object created with DcpJpegInit(2) functions
 * \param pimage_width  [OUT]
 * \param pimage_height [OUT]
 * \param pinput_components [OUT]
 * \param pin_color_space   [OUT]
 * \see CpJpegSetParams
 */
void DcpJpegGetParams( DcpJpeg * pDcpJpeg, unsigned int * pimage_width, 
        unsigned int * pimage_height,int * pinput_components, int * pin_color_space);

/*!\brief Decompresses image
 *
 * \param pDcpJpeg [IN] Pointer to object created with DcpJpegInit(2) functions
 * \param ppDecompress [OUT] Buffer to uncompressed data
 * \param pcbDecompress [OUT]
 * \return < 0 at the end or if error
 */
int DcpJpegDecode( DcpJpeg * pDcpJpeg, char **ppDecompress, int * pcbDecompress );

/*!\brief Creates the object used to handle compression
 *
 * \param WriteData [IN] Callback used to write the data
 * \param lParam  [IN] User data given to GetData and ErrorMessage
 * \param ErrorMessage [IN] Callback called when error
 * \return Pointer to a CpJpeg object
 */
CpJpeg * CpJpegInit( PFWriteData WriteData, void * lParam, PFErrorMessage ErrorMessage );

/*!\brief Destroy the compression object
 *
 * \param pCpJpeg [IN] Pointer to object created with CpJpegInit function
 */
void CpJpegDestroy( CpJpeg * pCpJpeg );

/*!\brief Sets image parameters
 *
 * \param pCpJpeg [IN] Pointer to object created with CpJpegInit function
 * \param image_width [IN]
 * \param image_height [IN]
 * \param input_components [IN]
 * \param in_color_space [IN]
 * \param nQuality [IN] The quality (1 to 100)
 * \see DcpJpegGetParams
 */
void CpJpegSetParams( CpJpeg * pCpJpeg, unsigned int image_width, 
	unsigned int image_height,int input_components, int in_color_space, int nQuality );

/*!\brief Compresses JPEG
 *
 * \param pCpJpeg [IN] Pointer to object created with CpJpegInit function
 * \param pData Data to compress
 * \param cbData Size of data
 * \return 0 if OK 
 */
int CpJpegEncode( CpJpeg * pCpJpeg, char *pData, int cbData );

/*!\brief Exemple  */
void TestJpeg( char *pszFileSource, char *pszFileDest, int nQuality, int fOnePass );


#ifdef __cplusplus
}
#endif


#endif	/*__DCPJPEG_H__*/
