/***************************************************************************
 * Wrapper to the jpeglib 6b
 * should be linked with the jpeglib-6b (http://www.ijg.org/)
 * To learn how to use it, look at function TestJpeg() and jpeg2jpeg
 * 
 * Sylvain MARECHAL - sylvain.marechal1@libertysurf.fr
 **************************************************************************/
#include "apr_buckets.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"                                                             
#include "dcpjpeg.c"


module AP_MODULE_DECLARE_DATA jpeg_module;

typedef struct jpeg_filter_config 
{
    int qualitylevel;
} jpeg_filter_config;

typedef struct jpeg_ctx 
{
    ap_filter_t * current_f;
    apr_bucket_brigade * current_bb;

    DcpJpeg * dcp_jpeg;
    CpJpeg * cp_jpeg;
    int init_done;
   
    int error;
    
    apr_bucket_brigade *bb;

    int read_size;
    int write_size;
} jpeg_ctx;


/***************************************************************************
 *
 **************************************************************************/
static const char *jpeg_set_qualitylevel( cmd_parms *cmd, void *dummy,
                                        const char *arg )
{
    jpeg_filter_config *c = ap_get_module_config( cmd->server->module_config,
                                                    &jpeg_module );
    int i;

    i = atoi(arg);

    if (i < 1 || i > 100)
        return "Quality Level must be between 1 and 100";

    c->qualitylevel = i;

    return NULL;
}

/***************************************************************************
 *
 **************************************************************************/
static const command_rec jpeg_filter_cmds[] =
{
    AP_INIT_TAKE1( "JpegQualityLevel", jpeg_set_qualitylevel, NULL, RSRC_CONF,
                  "Set the JPEG Quality Level (1-100)" ),
    {NULL}
};

/***************************************************************************
 *
 **************************************************************************/
static void *create_jpeg_server_config( apr_pool_t *p, server_rec *s )
{
    jpeg_filter_config *new =
    (jpeg_filter_config *) apr_pcalloc( p, sizeof(jpeg_filter_config) );

    new->qualitylevel = 5;
    return (void *) new;
}

/***************************************************************************
 *
 **************************************************************************/
void set_jpeg_ctx_current_params( jpeg_ctx * ctx, ap_filter_t * f, apr_bucket_brigade * bb )
{
    ctx->current_f = f;
    ctx->current_bb = bb;
}

/***************************************************************************
 *
 **************************************************************************/
static void read_jpeg_bucket_delete( void * lParam, void * lParam2 )
{
    apr_bucket *e = (apr_bucket *)lParam2;
    apr_bucket_delete(e);
}

/***************************************************************************
 *
 **************************************************************************/
static int read_jpeg( void * void_ctx, char ** ppBuffer, int * pcbBuffer,
        PFGetDataRelease * pGetDataRelease, void ** plParam2 )
{
    jpeg_ctx * ctx = (jpeg_ctx *)void_ctx;
    apr_bucket_brigade *bb = ctx->current_bb;
    ap_filter_t * f = ctx->current_f;
    int ret = -1;
    
    *pGetDataRelease = NULL;;

    /* Read data from current bucket */
    if( !APR_BRIGADE_EMPTY(bb) )
    {
        apr_bucket *e;

        e = APR_BRIGADE_FIRST(bb);
        if( APR_BUCKET_IS_EOS(e) ) 
        {
            ret = -1;
        }

        else if( APR_BUCKET_IS_FLUSH(e) ) 
        {
            ret = 0;
        }

        else
        {
            int rv;
            const char * data = NULL;
            int len = 0;
            /* read */
            rv = apr_bucket_read( e, &data, &len, 
                    /*APR_BLOCK_READ*/ APR_NONBLOCK_READ );
            *ppBuffer = (char **)data;
            *pcbBuffer = len;
            ret = len;
            ctx->read_size += len;
            if( APR_STATUS_IS_EAGAIN(rv) )
            {
                *ppBuffer = NULL;
                *pcbBuffer = 0;
            }
            else
            {
                /* delete buclet when data is copied */
                *pGetDataRelease = read_jpeg_bucket_delete;
                *plParam2 = e;
            }
        }
        if( pGetDataRelease == NULL )
        {
            apr_bucket_delete(e);
        }
    }
    else
    {
        *ppBuffer = NULL;
        *pcbBuffer = 0;
        ret = 0;
    }
 
    /* Return qtity of data readen, -1 if eof */
    return ret;
}

/***************************************************************************
 *
 **************************************************************************/
static int write_jpeg( void * void_ctx, char * pBuffer, int cbBuffer )
{
    apr_bucket * e;
    jpeg_ctx * ctx = (jpeg_ctx *)void_ctx;
    apr_bucket_brigade *bb = ctx->current_bb;
    ap_filter_t * f = ctx->current_f;
    request_rec * r = f->r;
    int ret = 0;

    ctx->write_size += cbBuffer;
    if( pBuffer == 0 && cbBuffer == 0 )
    {
        e = apr_bucket_eos_create( f->c->bucket_alloc );
    }
    else
    {
        e = apr_bucket_transient_create( pBuffer, cbBuffer, f->c->bucket_alloc);
    }
    APR_BRIGADE_INSERT_TAIL(ctx->bb, e);
    ap_pass_brigade(f->next, ctx->bb);
    return ret;
}

/***************************************************************************
 *
 **************************************************************************/
static void error_jpeg( void * void_ctx , char * pszBuffer )
{
    jpeg_ctx * ctx = (jpeg_ctx *)void_ctx;
    request_rec *r = ctx->current_f->r;
    ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r,
                   "JPEG error: "
                   "'%s' : URL %s",
                   pszBuffer, r->uri );
    ctx->error = 1;
}

/***************************************************************************
 *
 **************************************************************************/
static apr_status_t cleanup_jpeg_ctx( void * void_ctx )
{
    jpeg_ctx * ctx = (jpeg_ctx *)void_ctx;
    
    if( ctx->dcp_jpeg ) DcpJpegDestroy( ctx->dcp_jpeg );
    ctx->dcp_jpeg = NULL;
    if( ctx->cp_jpeg ) CpJpegDestroy( ctx->cp_jpeg );
    ctx->cp_jpeg = NULL;
    return 0;
}

/***************************************************************************
 *
 **************************************************************************/
static jpeg_ctx * create_jpeg_ctx( ap_filter_t  * f, jpeg_filter_config * config )
{
    jpeg_ctx * ctx = apr_pcalloc( f->r->pool, sizeof(*ctx) );
    if( ! ctx ) 
    {
        return ctx;
    }
    ctx->dcp_jpeg = DcpJpegInit2( read_jpeg, (void * )ctx, error_jpeg, 1 );
    ctx->cp_jpeg = CpJpegInit( write_jpeg, (void * )ctx, error_jpeg );
    DcpJpegInitScale( ctx->dcp_jpeg, 2, 1, 0 );
        
    ctx->bb = apr_brigade_create( f->r->pool, f->c->bucket_alloc );
    //ctx->buffer = apr_palloc( pool, c->bufferSize );

    apr_pool_userdata_set( ctx, "JPEG free pool callback", cleanup_jpeg_ctx, f->r->pool );

    return ctx;
}

/***************************************************************************
 *
 **************************************************************************/
static int is_last_bucket( jpeg_ctx * ctx )
{
    apr_bucket_brigade *bb = ctx->current_bb;
//    apr_bucket *e;

    if( !APR_BRIGADE_EMPTY(bb) ) return 0;
    return 1;
/*    e = APR_BRIGADE_FIRST(bb);
    if( APR_BUCKET_IS_EOS(e) ) return 1;
    return 0;*/
}

/***************************************************************************
 *
 **************************************************************************/
static apr_status_t process_jpeg( jpeg_ctx * ctx, jpeg_filter_config * config )
{
    char * dcp_buffer; int cb;
    do
    {
	int err;
	
	err = DcpJpegDecode( ctx->dcp_jpeg, &dcp_buffer, &cb );
	if( cb > 0 )
	{
	    if( ctx->init_done == 0 )
	    {
		unsigned int image_width;
		unsigned int image_height;
		int input_components;
		int in_color_space;
		
		DcpJpegGetParams( ctx->dcp_jpeg, &image_width, &image_height, 
                        &input_components, &in_color_space );
		CpJpegSetParams( ctx->cp_jpeg, image_width, image_height, 
                        input_components, in_color_space, config->qualitylevel );
		ctx->init_done = 1;
	    }
	    CpJpegEncode( ctx->cp_jpeg, dcp_buffer, cb );
	}
	if( err < 0 )
        {
	    break;
        }
    } while( cb > 0 || is_last_bucket(ctx) == 0 );
    return 0; 
}

/***************************************************************************
 *
 **************************************************************************/
static int is_jpeg_to_process( ap_filter_t *f, jpeg_filter_config * config )
{
    return 1;
}

/***************************************************************************
 *
 **************************************************************************/
static apr_status_t jpeg_out_filter(ap_filter_t *f, apr_bucket_brigade *bb)
{
    apr_bucket *e;
    request_rec *r = f->r;
    jpeg_ctx *ctx = f->ctx;
    jpeg_filter_config * config = ap_get_module_config(r->server->module_config,
            &jpeg_module);
    
    if (ctx == NULL) 
    {
        if( ! is_jpeg_to_process( f, config ) )
        {
            ap_remove_output_filter(f);
            return ap_pass_brigade(f->next, bb);
        }
        f->ctx = ctx = create_jpeg_ctx( f, config );

        apr_table_unset( f->r->headers_out, "Content-Length" );
    }
    
    set_jpeg_ctx_current_params( ctx, f, bb );
    process_jpeg( ctx, config );
    apr_brigade_cleanup(bb);
    if( ctx->error ) return APR_EGENERAL;
    return APR_SUCCESS;	
}

/***************************************************************************
 *
 **************************************************************************/
static void jpeg_register_hook(apr_pool_t *p)
{
    ap_register_output_filter( "JPEG", jpeg_out_filter, 
            NULL, /*AP_FTYPE_CONTENT_SET*/ AP_FTYPE_RESOURCE );
}


/***************************************************************************
 *
 **************************************************************************/
module AP_MODULE_DECLARE_DATA jpeg_module = 
{
    STANDARD20_MODULE_STUFF,
    NULL,		        /* create per-directory config structure */
    NULL,        		/* merge per-directory config structures */
    create_jpeg_server_config,	/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    jpeg_filter_cmds,		/* command apr_table_t */
    jpeg_register_hook		/* register hooks */
};

